/*
 *  SerialUSB.cpp
 *  SerialUSB
 *
 *  Created by Tom Saxton on 4/18/07.
 *  Copyright 2007 Idle Loop Software Design, LLC. All rights reserved.
 *
 *  Permission granted for anyone to adapt or use this code without restriction,
 *  however no guarantee or warranty of any sort offered. Use at your own risk.
 *
 */

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/serial/IOSerialKeys.h>
#include <IOKit/IOBSD.h>
#include <IOKit/usb/IOUSBLib.h>

#include <string.h>

#include "SerialUSB.h"

// if the specified object, or any of its kIOServicePlane parent objects, is a USB device,
// return YES and fill in the Vendor and Product ID values,
// otherwise return NO
Boolean FIsUSBDevice(io_object_t obj, SInt32 *pidVendor, SInt32 *pidProduct)
{
	// we need to release all of the objects we fetch,
	// but not the object that was originally passed in.
	io_object_t objRelease = NULL;

	// default return value
	Boolean fIsUSB = 0;

	while(obj)
	{
		char szT[100];
	
		// is this an IOUSBDevice?
		if (FGetIOObjectPropertySz(obj, CFSTR(kIOProviderClassKey), szT, sizeof(szT))
			&& strcmp(szT, kIOUSBDeviceClassName) == 0
			)
		{
			// if so, get the vendor and product IDs
			if (FGetIOObjectPropertySInt32(obj, CFSTR(kUSBVendorID), pidVendor)
				&& FGetIOObjectPropertySInt32(obj, CFSTR(kUSBProductID), pidProduct))
			{
				// success
				fIsUSB = 1;
				break;
			}
		}

		// get the object's parent
		obj = IOObjectGetParent(obj);
		
		// if we fetched it, release the previous object now
		if (objRelease)
		{
			IOObjectRelease(objRelease);
			objRelease = NULL;
		}
		
		// set the object we just fetched to be released
		objRelease = obj;
	}

	// take care of any pending release
	if (objRelease)
	{
		IOObjectRelease(objRelease);
		objRelease = NULL;
	}
	
	return fIsUSB;
}

// get the kIOServicePlane parent of the specified object
// the parent object must be released by the caller
io_object_t IOObjectGetParent(io_object_t obj)
{
	kern_return_t kernResult;
	io_object_t objParent;
	
	if ((kernResult = IORegistryEntryGetParentEntry(obj, kIOServicePlane, &objParent)) != KERN_SUCCESS)
		return NULL;

	return objParent;
}

// get the specified property of the object as a string
Boolean FGetIOObjectPropertySz(io_object_t obj, CFStringRef strrefProperty, char psz[], int cb)
{
	Boolean fResult = 1;

	// initialize result to empty string
	if (psz != NULL && cb >= 1)
		psz[0] = 0;

	// get the property entry
	CFStringRef strrefT = (CFStringRef)IORegistryEntryCreateCFProperty(obj,
												strrefProperty,
												kCFAllocatorDefault,
												0);

	// bail on failure
	if (strrefT == NULL)
		return 0;

	// convert to C string
	if (psz != NULL)
		fResult = CFStringGetCString(strrefT, psz, cb, kCFStringEncodingASCII);

	CFRelease(strrefT);
	
	return fResult;
}

// get the specified property of the object as an SInt32
Boolean FGetIOObjectPropertySInt32(io_object_t obj, CFStringRef strrefProperty, SInt32 *pnum)
{
	Boolean fResult = 1;
	
	// default return value
	if (pnum != NULL)
		*pnum = 0;

	// get the property entry
	CFNumberRef numref = (CFNumberRef)IORegistryEntryCreateCFProperty(obj,
											strrefProperty,
											kCFAllocatorDefault,
											0);

	// bail on failure
	if (numref == NULL)
		return 0;

	// fill in the number
	if (pnum != NULL)
		fResult = CFNumberGetValue(numref, kCFNumberSInt32Type, pnum);

	CFRelease(numref);

	return fResult;
}
