/*
 *  main.cpp
 *  SerialUSB
 *
 *  Created by Tom Saxton on 4/18/07.
 *  Copyright 2007 Idle Loop Software Design, LLC.
 *
 *  Permission granted for anyone to adapt or use this code without restriction,
 *  however no guarantee or warranty of any sort offered. Use at your own risk.
 *
 *  NOTE: the function FindSerialPorts is from sample code by Apple, Inc, although
 *  it may have been changed. ListSerialPorts is also derived from Apple sample code.
 *  Be sure to consult Apple's sample code as the definitive resource, don't blame them
 *  for my bugs.
 */

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/serial/IOSerialKeys.h>
#include <IOKit/IOBSD.h>
#include <IOKit/usb/IOUSBLib.h>

#include "SerialUSB.h"

kern_return_t FindSerialPorts(io_iterator_t *matchingServices);
void ListSerialPorts();

int main (int argc, char * const argv[])
{
	ListSerialPorts();
    return 0;
}

void ListSerialPorts()
{
    kern_return_t kernResult;
    io_iterator_t iteratorSerialPort;

	kernResult = FindSerialPorts(&iteratorSerialPort);
	if (kernResult == KERN_SUCCESS)
	{
		io_object_t objSerialPort;
		IOIteratorReset(iteratorSerialPort);

		while ((objSerialPort = IOIteratorNext(iteratorSerialPort)) != 0)
		{
			char szPort[4*1024];
			SInt32 idVendor, idProduct;

			// get the port name for the object
			if (!FGetIOObjectPropertySz(objSerialPort, CFSTR(kIOCalloutDeviceKey), szPort, sizeof(szPort)))
			{
				fprintf(stderr, "ERROR: object 0x%08X has no kIOCalloutDeviceKey property!\n", objSerialPort);
				continue;
			}
			
			// print the object reference and the device file path
			printf("0x%08X: %s", objSerialPort, szPort);
			
			// if it's a USB thing, print the vendor and product IDs
			if (FIsUSBDevice(objSerialPort, &idVendor, &idProduct))
				printf(", Vendor ID = 0x%04lX, Product ID = 0x%04lX", idVendor, idProduct);
			
			// end the line
			printf("\n");

			// Release the io_service_t now that we are done with it.
			IOObjectRelease(objSerialPort);
		}
	}
    
    IOObjectRelease(iteratorSerialPort);	// Release the iterator.
}

// Returns an iterator across all known modems. Caller is responsible for
// releasing the iterator when iteration is complete.
kern_return_t FindSerialPorts(io_iterator_t *matchingServices)
{
    kern_return_t		kernResult; 
    mach_port_t			masterPort;
    CFMutableDictionaryRef	classesToMatch;

/*! @function IOMasterPort
    @abstract Returns the mach port used to initiate communication with IOKit.
    @discussion Functions that don't specify an existing object require the IOKit master port to be passed. This function obtains that port.
    @param bootstrapPort Pass MACH_PORT_NULL for the default.
    @param masterPort The master port is returned.
    @result A kern_return_t error code. */

    kernResult = IOMasterPort(MACH_PORT_NULL, &masterPort);
    if (KERN_SUCCESS != kernResult)
    {
		printf("IOMasterPort returned %d\n", kernResult);
		goto exit;
    }
        
/*! @function IOServiceMatching
    @abstract Create a matching dictionary that specifies an IOService class match.
    @discussion A very common matching criteria for IOService is based on its class. IOServiceMatching will create a matching dictionary that specifies any IOService of a class, or its subclasses. The class is specified by C-string name.
    @param name The class name, as a const C-string. Class matching is successful on IOService's of this class or any subclass.
    @result The matching dictionary created, is returned on success, or zero on failure. The dictionary is commonly passed to IOServiceGetMatchingServices or IOServiceAddNotification which will consume a reference, otherwise it should be released with CFRelease by the caller. */

    // Serial devices are instances of class IOSerialBSDClient
    classesToMatch = IOServiceMatching(kIOSerialBSDServiceValue);
    if (classesToMatch == NULL)
    {
        printf("IOServiceMatching returned a NULL dictionary.\n");
    }
    else {
/*!
	@function CFDictionarySetValue
	Sets the value of the key in the dictionary.
	@param theDict The dictionary to which the value is to be set. If this
		parameter is not a valid mutable CFDictionary, the behavior is
		undefined. If the dictionary is a fixed-capacity dictionary and
		it is full before this operation, and the key does not exist in
		the dictionary, the behavior is undefined.
	@param key The key of the value to set into the dictionary. If a key 
		which matches this key is already present in the dictionary, only
		the value is changed ("add if absent, replace if present"). If
		no key matches the given key, the key-value pair is added to the
		dictionary. If added, the key is retained by the dictionary,
		using the retain callback provided
		when the dictionary was created. If the key is not of the sort
		expected by the key retain callback, the behavior is undefined.
	@param value The value to add to or replace into the dictionary. The value
		is retained by the dictionary using the retain callback provided
		when the dictionary was created, and the previous value if any is
		released. If the value is not of the sort expected by the
		retain or release callbacks, the behavior is undefined.
*/
        CFDictionarySetValue(classesToMatch,
                             CFSTR(kIOSerialBSDTypeKey),
                             CFSTR(kIOSerialBSDAllTypes));

        // Each serial device object has a property with key
        // kIOSerialBSDTypeKey and a value that is one of kIOSerialBSDAllTypes,
        // kIOSerialBSDModemType, or kIOSerialBSDRS232Type. You can experiment with the
        // matching by changing the last parameter in the above call to CFDictionarySetValue.
        
        // As shipped, this sample is only interested in modems,
        // so add this property to the CFDictionary we're matching on. 
        // This will find devices that advertise themselves as modems,
        // such as built-in and USB modems. However, this match won't find serial modems.
    }
    
    /*! @function IOServiceGetMatchingServices
        @abstract Look up registered IOService objects that match a matching dictionary.
        @discussion This is the preferred method of finding IOService objects currently registered by IOKit. IOServiceAddNotification can also supply this information and install a notification of new IOServices. The matching information used in the matching dictionary may vary depending on the class of service being looked up.
        @param masterPort The master port obtained from IOMasterPort().
        @param matching A CF dictionary containing matching information, of which one reference is consumed by this function. IOKitLib can contruct matching dictionaries for common criteria with helper functions such as IOServiceMatching, IOOpenFirmwarePathMatching.
        @param existing An iterator handle is returned on success, and should be released by the caller when the iteration is finished.
        @result A kern_return_t error code. */

    kernResult = IOServiceGetMatchingServices(masterPort, classesToMatch, matchingServices);    
    if (KERN_SUCCESS != kernResult)
    {
        printf("IOServiceGetMatchingServices returned %d\n", kernResult);
		goto exit;
    }
        
exit:
    return kernResult;
}
